/*
 * Print the contents of an i40e_tx_desc_t.cmd_type_offset_bsz.
 */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>

#define I40E_TXD_QW1_DTYPE_SHIFT	0
#define I40E_TXD_QW1_DTYPE_MASK		(0xFUL << I40E_TXD_QW1_DTYPE_SHIFT)

enum i40e_tx_desc_dtype_value {
	I40E_TX_DESC_DTYPE_DATA		= 0x0,
	I40E_TX_DESC_DTYPE_NOP		= 0x1, /* same as Context desc */
	I40E_TX_DESC_DTYPE_CONTEXT	= 0x1,
	I40E_TX_DESC_DTYPE_FCOE_CTX	= 0x2,
	I40E_TX_DESC_DTYPE_FILTER_PROG	= 0x8,
	I40E_TX_DESC_DTYPE_DDP_CTX	= 0x9,
	I40E_TX_DESC_DTYPE_FLEX_DATA	= 0xB,
	I40E_TX_DESC_DTYPE_FLEX_CTX_1	= 0xC,
	I40E_TX_DESC_DTYPE_FLEX_CTX_2	= 0xD,
	I40E_TX_DESC_DTYPE_DESC_DONE	= 0xF
};

#define I40E_TXD_QW1_CMD_SHIFT	4
#define I40E_TXD_QW1_CMD_MASK	(0x3FFUL << I40E_TXD_QW1_CMD_SHIFT)

enum i40e_tx_desc_cmd_bits {
	I40E_TX_DESC_CMD_EOP			= 0x0001,
	I40E_TX_DESC_CMD_RS			= 0x0002,
	I40E_TX_DESC_CMD_ICRC			= 0x0004,
	I40E_TX_DESC_CMD_IL2TAG1		= 0x0008,
	I40E_TX_DESC_CMD_DUMMY			= 0x0010,
	I40E_TX_DESC_CMD_IIPT_NONIP		= 0x0000, /* 2 BITS */
	I40E_TX_DESC_CMD_IIPT_IPV6		= 0x0020, /* 2 BITS */
	I40E_TX_DESC_CMD_IIPT_IPV4		= 0x0040, /* 2 BITS */
	I40E_TX_DESC_CMD_IIPT_IPV4_CSUM		= 0x0060, /* 2 BITS */
	I40E_TX_DESC_CMD_FCOET			= 0x0080,
	I40E_TX_DESC_CMD_L4T_EOFT_UNK		= 0x0000, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_TCP		= 0x0100, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_SCTP		= 0x0200, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_UDP		= 0x0300, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_EOF_N		= 0x0000, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_EOF_T		= 0x0100, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_EOF_NI	= 0x0200, /* 2 BITS */
	I40E_TX_DESC_CMD_L4T_EOFT_EOF_A		= 0x0300, /* 2 BITS */
};

#define I40E_TXD_QW1_OFFSET_SHIFT	16
#define I40E_TXD_QW1_OFFSET_MASK	(0x3FFFFULL << \
					 I40E_TXD_QW1_OFFSET_SHIFT)

enum i40e_tx_desc_length_fields {
	/* Note: These are predefined bit offsets */
	I40E_TX_DESC_LENGTH_MACLEN_SHIFT	= 0, /* 7 BITS */
	I40E_TX_DESC_LENGTH_IPLEN_SHIFT		= 7, /* 7 BITS */
	I40E_TX_DESC_LENGTH_L4_FC_LEN_SHIFT	= 14 /* 4 BITS */
};

#define I40E_TXD_QW1_MACLEN_MASK (0x7FUL << I40E_TX_DESC_LENGTH_MACLEN_SHIFT)
#define I40E_TXD_QW1_IPLEN_MASK  (0x7FUL << I40E_TX_DESC_LENGTH_IPLEN_SHIFT)
#define I40E_TXD_QW1_L4LEN_MASK  (0xFUL << I40E_TX_DESC_LENGTH_L4_FC_LEN_SHIFT)
#define I40E_TXD_QW1_FCLEN_MASK  (0xFUL << I40E_TX_DESC_LENGTH_L4_FC_LEN_SHIFT)

#define I40E_TXD_QW1_TX_BUF_SZ_SHIFT	34
#define I40E_TXD_QW1_TX_BUF_SZ_MASK	(0x3FFFULL << \
					 I40E_TXD_QW1_TX_BUF_SZ_SHIFT)

/*
 * Context descriptor masks.
 */
#define I40E_TXD_CTX_QW1_CMD_SHIFT	4
#define I40E_TXD_CTX_QW1_CMD_MASK	(0xFFFFUL << I40E_TXD_CTX_QW1_CMD_SHIFT)

enum i40e_tx_ctx_desc_cmd_bits {
	I40E_TX_CTX_DESC_TSO		= 0x01,
	I40E_TX_CTX_DESC_TSYN		= 0x02,
	I40E_TX_CTX_DESC_IL2TAG2	= 0x04,
	I40E_TX_CTX_DESC_IL2TAG2_IL2H	= 0x08,
	I40E_TX_CTX_DESC_SWTCH_NOTAG	= 0x00,
	I40E_TX_CTX_DESC_SWTCH_UPLINK	= 0x10,
	I40E_TX_CTX_DESC_SWTCH_LOCAL	= 0x20,
	I40E_TX_CTX_DESC_SWTCH_VSI	= 0x30,
	I40E_TX_CTX_DESC_SWPE		= 0x40
};

#define I40E_TXD_CTX_QW1_TSO_LEN_SHIFT	30
#define I40E_TXD_CTX_QW1_TSO_LEN_MASK	(0x3FFFFULL << \
					 I40E_TXD_CTX_QW1_TSO_LEN_SHIFT)

#define I40E_TXD_CTX_QW1_MSS_SHIFT	50
#define I40E_TXD_CTX_QW1_MSS_MASK	(0x3FFFULL << \
					 I40E_TXD_CTX_QW1_MSS_SHIFT)

#define I40E_TXD_CTX_QW1_VSI_SHIFT	50
#define I40E_TXD_CTX_QW1_VSI_MASK	(0x1FFULL << I40E_TXD_CTX_QW1_VSI_SHIFT)

#define	SIZE	1024

void
print_ctx_desc(unsigned int idx, uint64_t bits)
{
	uint64_t tbits = 0;
	char cmd[SIZE];
	char *type = "CTX";
	unsigned int mss;
	unsigned int tlen;

	snprintf(cmd, SIZE, "");

	tbits = bits & I40E_TXD_CTX_QW1_CMD_MASK;
	tbits = tbits >> I40E_TXD_CTX_QW1_CMD_SHIFT;

	if (tbits & I40E_TX_CTX_DESC_TSO)
		snprintf(cmd, SIZE, "%s|", "TSO");

	if (tbits & I40E_TX_CTX_DESC_TSYN)
		snprintf(cmd, SIZE, "%s%s", cmd, "TSYN|");

	if (tbits & I40E_TX_CTX_DESC_IL2TAG2)
		snprintf(cmd, SIZE, "%s%s", cmd, "IL2TAG2|");

	if (tbits & I40E_TX_CTX_DESC_IL2TAG2_IL2H)
		snprintf(cmd, SIZE, "%s%s", cmd, "IL2TAG2_IL2H|");

	if ((tbits & I40E_TX_CTX_DESC_SWTCH_VSI) ==
	    I40E_TX_CTX_DESC_SWTCH_NOTAG)
		snprintf(cmd, SIZE, "%s%s", cmd, "NOTAG|");

	if ((tbits & I40E_TX_CTX_DESC_SWTCH_VSI) ==
	    I40E_TX_CTX_DESC_SWTCH_UPLINK)
		snprintf(cmd, SIZE, "%s%s", cmd, "UPLINK|");

	if ((tbits & I40E_TX_CTX_DESC_SWTCH_VSI) ==
	    I40E_TX_CTX_DESC_SWTCH_LOCAL)
		snprintf(cmd, SIZE, "%s%s", cmd, "LOCAL|");

	if ((tbits & I40E_TX_CTX_DESC_SWTCH_VSI) ==
	    I40E_TX_CTX_DESC_SWTCH_VSI)
		snprintf(cmd, SIZE, "%s%s", cmd, "VSI|");

	tbits = bits & I40E_TXD_CTX_QW1_TSO_LEN_MASK;
	tlen = tbits >> I40E_TXD_CTX_QW1_TSO_LEN_SHIFT;

	tbits = bits & I40E_TXD_CTX_QW1_MSS_MASK;
	mss = tbits >> I40E_TXD_CTX_QW1_MSS_SHIFT;

	printf("%-4u 0x%-16llx %-4s %-24s %-4u %-4u\n",
	    idx, bits, type, cmd, tlen, mss);
}

void
print_data_desc(unsigned int idx, uint64_t bits)
{
	uint64_t tbits = 0;
	uint32_t offset = 0;
	uint16_t maclen, iplen, l4len, bufsz;
	char cmd[SIZE];
	char *type;

	snprintf(cmd, SIZE, "");
	tbits = bits & I40E_TXD_QW1_DTYPE_MASK;

	if (tbits  == I40E_TX_DESC_DTYPE_DATA)
		type = "DATA";
	else if (tbits == I40E_TX_DESC_DTYPE_NOP)
		type = "CTX";
	else if (tbits == I40E_TX_DESC_DTYPE_DESC_DONE)
		type = "DONE";
	else
		type = "OTH";

	tbits = bits & I40E_TXD_QW1_CMD_MASK;
	tbits = tbits >> I40E_TXD_QW1_CMD_SHIFT;

	if (tbits & I40E_TX_DESC_CMD_EOP)
		snprintf(cmd, SIZE, "EOP|");

	if (tbits & I40E_TX_DESC_CMD_RS)
		snprintf(cmd, SIZE, "%s%s", cmd, "RS|");

	if (tbits & I40E_TX_DESC_CMD_IL2TAG1)
		snprintf(cmd, SIZE, "%s%s", cmd, "TAG|");

	if (tbits & I40E_TX_DESC_CMD_DUMMY)
		snprintf(cmd, SIZE, "%s%s", cmd, "DUMMY|");

	if ((tbits & I40E_TX_DESC_CMD_IIPT_IPV4_CSUM) ==
	    I40E_TX_DESC_CMD_IIPT_NONIP)
		snprintf(cmd, SIZE, "%s%s", cmd, "NONIP|");

	if ((tbits & I40E_TX_DESC_CMD_IIPT_IPV4_CSUM) ==
	    I40E_TX_DESC_CMD_IIPT_IPV6)
		snprintf(cmd, SIZE, "%s%s", cmd, "IPV6|");

	if ((tbits & I40E_TX_DESC_CMD_IIPT_IPV4_CSUM) ==
	    I40E_TX_DESC_CMD_IIPT_IPV4)
		snprintf(cmd, SIZE, "%s%s", cmd, "IPV4|");

	if ((tbits & I40E_TX_DESC_CMD_IIPT_IPV4_CSUM) ==
	    I40E_TX_DESC_CMD_IIPT_IPV4_CSUM)
		snprintf(cmd, SIZE, "%s%s", cmd, "IPV4CSUM|");

	if ((tbits & I40E_TX_DESC_CMD_L4T_EOFT_UDP) ==
	    I40E_TX_DESC_CMD_L4T_EOFT_UNK)
		snprintf(cmd, SIZE, "%s%s", cmd, "UNK|");

	if ((tbits & I40E_TX_DESC_CMD_L4T_EOFT_UDP) ==
	    I40E_TX_DESC_CMD_L4T_EOFT_TCP)
		snprintf(cmd, SIZE, "%s%s", cmd, "TCP|");

	if ((tbits & I40E_TX_DESC_CMD_L4T_EOFT_UDP) ==
	    I40E_TX_DESC_CMD_L4T_EOFT_SCTP)
		snprintf(cmd, SIZE, "%s%s", cmd, "SCTP|");

	if ((tbits & I40E_TX_DESC_CMD_L4T_EOFT_UDP) ==
	    I40E_TX_DESC_CMD_L4T_EOFT_UDP)
		snprintf(cmd, SIZE, "%s%s", cmd, "UDP|");

	tbits = bits & I40E_TXD_QW1_OFFSET_MASK;
	tbits = tbits >> I40E_TXD_QW1_OFFSET_SHIFT;
	offset = tbits;

	/*
	 * MAC len defined in 2-byte words, others defined in 4-byte
	 * double-words.
	 */
	maclen = ((offset & I40E_TXD_QW1_MACLEN_MASK) >>
	    I40E_TX_DESC_LENGTH_MACLEN_SHIFT) * 2;
	iplen = ((offset & I40E_TXD_QW1_IPLEN_MASK) >>
	    I40E_TX_DESC_LENGTH_IPLEN_SHIFT) * 4;
	l4len = ((offset & I40E_TXD_QW1_L4LEN_MASK) >>
	    I40E_TX_DESC_LENGTH_L4_FC_LEN_SHIFT) * 4;

	tbits = bits & I40E_TXD_QW1_TX_BUF_SZ_MASK;
	bufsz = tbits >> I40E_TXD_QW1_TX_BUF_SZ_SHIFT;

	printf("%-4u 0x%-16llx %-4s %-24s %-4u %-4u %-4u %-6u\n",
	    idx, bits, type, cmd, maclen, iplen, l4len, bufsz);
}

int
main(int argc, char *argv[])
{
	uint64_t bits;
	char *line;
	size_t len = SIZE;
	unsigned int i = 0;
	FILE *fp;
	ssize_t read;

	if (argc != 2) {
		fprintf(stderr, "must provide file with hex contents");
		return (-1);
	}

	if ((line = malloc(SIZE)) == NULL) {
		fprintf(stderr, "Failed to malloc line buffer\n");
		return (1);
	}

	printf("%-4s %-18s %-4s %-24s %-4s %-4s %-4s %-6s\n",
	    "IDX", "VALUE", "TYPE", "CMD", "MAC", "IP", "L4", "BUFSZ");

	fp = fopen(argv[1], "r");
	if (fp == NULL) {
		fprintf(stderr, "Failed to open file: %s (%d)\n", argv[1],
		    errno);
		return (1);
	}

	while ((read = getline(&line, &len, fp)) != -1) {
		uint64_t tbits;

		bits = strtoull(line, NULL, 16);
		tbits = bits & I40E_TXD_QW1_DTYPE_MASK;

		if (tbits == I40E_TX_DESC_DTYPE_NOP)
			print_ctx_desc(i, bits);
		else
			print_data_desc(i, bits);

		i++;
	}

	fclose(fp);
	free(line);
	return (0);
}
